import math
import random
import cv2
from skimage.metrics import structural_similarity as ssim
from ....DeviceState import ScreenshotState
from ....DeviceControl import TuyaControl, LightControl


class TuyaLight(ScreenshotState, TuyaControl, LightControl):
    """
    Generic Tuya light bulb.
    """

    # Default filename for the OFF screenshot
    FILENAME_SCREENSHOT_OFF = "off.png"
    # SSIM thresholds
    SSIM_DIFF_THRESHOLD = 0.99      # State change
    OFF_SSIM_DIFF_THRESHOLD = 0.95  # Check if light is off

    ### Screen coordinates
    # Device controls
    device_x = 345.6
    device_y = 590.4
    ## Events
    # Toggle
    x = 360
    y = 1281.6
    # Brightness
    x_left_gauge  = 164
    x_right_gauge = 580
    y_brightness  = 936
    # Color
    x_color_center = 360
    y_color_center = 590.4
    y_color_ring   = 400


    def __init__(self, ipv4: str, **kwargs) -> None:
        """
        Constructor.
        Initializes the Tuya light bulb with its IP address,
        and computes the grayscale array of the bulb's OFF screenshot.

        Args:
            ipv4 (str): IP address of the Tuya light bulb.
            kwargs (dict): device-specific additional parameters,
                           including the path to the OFF screenshot.
        """
        # Call parent constructor
        super().__init__(ipv4, **kwargs)

        # Compute gray array of streaming screenshot
        image_off = cv2.imread(kwargs.get("path_screenshot_off", TuyaLight.FILENAME_SCREENSHOT_OFF))
        self.image_off_gray = cv2.cvtColor(image_off, cv2.COLOR_BGR2GRAY)
    

    def do_set_color(self) -> None:
        """
        Randomly set the color of the Tuya light.
        """
        # Generate random position on the color ring
        radius = self.y_color_center - self.y_color_ring
        angle_deg = random.randint(0, 360)
        angle_rad = angle_deg * math.pi / 180
        x = self.x_color_center + radius * math.cos(angle_rad)
        y = self.y_color_center + radius * math.sin(angle_rad)
        # Set the color
        self.get_phone().shell(f"input tap {x} {y}")
    

    def is_off(self):
        """
        Check if the light is off.
        """
        # Take screenshot and convert to grayscale
        state = self.get_state()
        screenshot_gray = cv2.cvtColor(state, cv2.COLOR_BGR2GRAY)

        # Compute SSIM with OFF screenshot
        try:
            score = ssim(self.image_off_gray, screenshot_gray, full=False)
            return score > self.OFF_SSIM_DIFF_THRESHOLD
        except ValueError:
            return False



from ....DeviceState import CameraScreenshotState
from ....DeviceControl import CameraControl


class TapoCamera(CameraScreenshotState, CameraControl):
    """
    Tapo camera (C200).
    """

    ### Class attributes
    # Android package
    android_package = "com.tplink.iot"
    # SSIM threshold below which images are considered different
    SSIM_DIFF_THRESHOLD = 0.95
    ## Screen coordinates
    # Stream event
    x_start = 187.2
    y_start = 446.4
    x_stop = 50.4
    y_stop = 100.8



from typing import Awaitable
from tapo import PlugEnergyMonitoringHandler
from .TapoState import TapoState


class TapoPlugState(TapoState):
    """
    Class to represent the state of a Tapo plug.
    """

    def __init__(self, ipv4: str, **kwargs) -> None:
        """
        Constructor.
        Initializes the Tapo device with its IP address,
        and creates a Tapo API client.

        Args:
            ipv4 (str): The Tapo device's IPv4 address.
            kwargs (dict): device-specific additional parameters,
                           including the TP-Link account credentials.
        """
        # Superclass (TapoState) constructor
        super().__init__(ipv4)

        # Tapo plug connector
        self.device: PlugEnergyMonitoringHandler = None


    async def _async_get_state(self) -> Awaitable[bool]:
        """
        Asynchronously get the state of the Tapo plug.

        Returns:
            Awaitable[bool]: async function which returns True if the plug is on, False otherwise.
        """
        if self.device is None:
            # Create a Tapo plug device handler
            self.device = await self.client.p110(self.ipv4)

        info = await self.device.get_device_info()
        return info.device_on



from .SmartThingsControl import SmartThingsControl
from .PlugControl import PlugControl


class SmartThingsPlugControl(SmartThingsControl, PlugControl):
    """
    Abstract class to control a generic plug through SmartThings.
    """

    # Toggle coordinates
    x = 961
    y = 436



import time
from miio import ChuangmiCamera
from ....DeviceControl import CameraControl


class XiaomiCamera(CameraControl, ChuangmiCamera):
    """
    Xiaomi camera (Chuangmi IPC019).
    """

    ### Class variables
    # Android package name
    android_package = "com.xiaomi.smarthome"
    ## Screen coordinates
    # Zone button
    x_zone = 345.6
    y_zone = 201.6
    # Camera control button
    x_camera = 619.2
    y_camera = 331.2
    # Stream event
    x_start = 100.8
    y_start = 489.6


    def __init__(self, ipv4: str, **kwargs) -> None:
        """
        Constructor.
        Initialize the Xiaomi camera (Chuangmi IPC019) with its IP address.

        Args:
            ipv4 (str): The device's IPv4 address.
            kwargs (dict): device-specific additional parameters, including device API token.
        """
        ChuangmiCamera.__init__(self, ip=ipv4, token=kwargs.get("token", ""))
        self._was_last_stream_successful = False
    

    def start_app(self) -> None:
        """
        Start the Mi Home app on the phone,
        then switch to the zone control screen,
        then to the camera's stream.

        Raises:
            IndexError: If no adb device is found.
        """
        # Start the Mi Home app
        phone = self.get_phone()
        phone.shell(f"monkey -p {XiaomiCamera.android_package} -c android.intent.category.LAUNCHER 1")
        time.sleep(10)
        # Switch to the camera control screen
        phone.shell(f"input tap {self.x_zone} {self.y_zone}")
        phone.shell(f"input tap {self.x_camera} {self.y_camera}")


    def start_stream(self) -> None:
        """
        Start the camera's video stream.
        Overrides the method from the `CameraControl` class.
        """
        phone = self.get_phone()
        phone.shell(f"input tap {self.x_start} {self.y_start}")
        phone.shell(f"input tap {self.x_start} {self.y_start}")
    

    def do_stop_stream(self) -> None:
        """
        Stop the camera's video stream.
        Overrides the method from the `CameraControl` class.
        """
        self.start_stream()
    

    def get_state(self) -> bool:
        """
        Get the state of the camera.

        Returns:
            bool: True if the camera is on, False otherwise.
        """
        return self.status().power



from typing import Any
import importlib


def init_device(name: str, ipv4: str, **kwargs) -> Any:
    """
    Factory method to create a device object.

    Args:
        name (str): The name of the device.
        ipv4 (str): The device's IPv4 address.
        kwargs (dict): device-specific additional parameters.
    Returns:
        Any: The device object.
    """
    package_parts = importlib.import_module(__name__).__name__.split(".")
    package_name = f"{package_parts[0]}.{package_parts[1]}"
    package = importlib.import_module(package_name)
    cls = getattr(package, name)
    return cls(ipv4, **kwargs)


def close_device(device: Any) -> None:
    """
    Close the device connector.

    Args:
        device (Any): The device object.
    """
    try:
        device.close()
    except AttributeError:
        # If the device does not have a stop method, we can ignore it
        pass



import random
from .DeviceControl import DeviceControl


class LightControl(DeviceControl):
    """
    Class to control a light bulb device.
    """


    def toggle(self) -> None:
        """
        Perform the toggle event on the light.
        """
        self.get_phone().shell(f"input tap {self.x} {self.y}")


    def _set_light_attr(self, y: int) -> None:
        """
        Randomly set an attribute of the light
        (brightness or color).
        
        Args:
            y (int): The y-coordinate of the corresponding gauge.
        """
        x = random.randint(self.x_left_gauge, self.x_right_gauge)
        self.get_phone().shell(f"input tap {x} {y}")

    
    def do_set_brightness(self) -> None:
        """
        Randomly set the brightness of the light.
        """
        self._set_light_attr(self.y_brightness)

    
    def set_brightness(self) -> None:
        """
        Template method to randomly set the brightness of the light.
        The child class must implement the concrete method `is_on`,
        which checks if the light is on.
        """
        # If light is off, turn it on
        if self.is_off():
            self.toggle()
        
        self.do_set_brightness()
    

    def do_set_color(self) -> None:
        """
        Randomly set the color of the light.
        """
        self._set_light_attr(self.y_color)

    
    def set_color(self) -> None:
        """
        Template method to randomly set the color of the light.
        The child class must implement the concrete method `is_on`,
        which checks if the light is on.
        """
        # If light is off, turn it on
        if self.is_off():
            self.toggle()
        
        self.do_set_color()



from .DeviceControl import DeviceControl


class PlugControl(DeviceControl):
    """
    Class to control a plug device.
    """


    def toggle(self) -> None:
        """
        Perform the toggle action on the plug.
        """
        self.get_phone().shell(f"input tap {self.x} {self.y}")



from .TapoControl import TapoControl
from .PlugControl import PlugControl


class TapoPlugControl(TapoControl, PlugControl):
    """
    Abstract class to control a generic plug through the Tapo app.
    """

    # Toggle coordinates
    x = 940
    y = 824



import time
from .DeviceControl import DeviceControl


class TuyaControl(DeviceControl):
    """
    Abstract class to control a generic device through the Tuya app.
    """

    # Metadata
    android_package = "com.tuya.smart"
    version = "3.3"


    def start_app(self) -> None:
        """
        Start the Tuya app on the phone.

        Raises:
            IndexError: If no adb device is found.
        """
        phone = self.get_phone()
        # Start Tapo app
        phone.shell(f"monkey -p {TuyaControl.android_package} -c android.intent.category.LAUNCHER 1")
        time.sleep(10)
        # Open device controls
        phone.shell(f"input tap {self.device_x} {self.device_y}")



import cv2
from skimage.metrics import structural_similarity as ssim
from .CameraState import CameraState


class CameraScreenshotState(CameraState):
    """
    Abstract class for the state of a camera device,
    which uses screenshot-based event validation.
    """

    FILENAME_SCREENSHOT_STREAM = "stream.png"


    def __init__(self, ipv4: str, **kwargs) -> None:
        """
        Constructor.
        Initialize the camera device with its IP address.

        Args:
            ipv4 (str): The device's IPv4 address.
            kwargs (dict): device-specific additional parameters,
                           including the path to the streaming screenshot.
        """
        # Call parent constructor
        super().__init__(ipv4, **kwargs)

        # Compute gray array of streaming screenshot
        stream_image = cv2.imread(kwargs.get("path_screenshot_stream", CameraScreenshotState.FILENAME_SCREENSHOT_STREAM))
        self.gray_stream_image = cv2.cvtColor(stream_image, cv2.COLOR_BGR2GRAY)


    def get_state(self) -> bool:
        """
        Get the state of the camera device,
        i.e. if it is currently streaming,
        by taking a screenshot of its app.

        Returns:
            bool: True if the camera is currently streaming, False otherwise.
        """
        # Take screenshot and convert to grayscale
        screenshot_array = self.screenshot()
        gray_screenshot = cv2.cvtColor(screenshot_array, cv2.COLOR_BGR2GRAY)

        # Compute SSIM
        try:
            score = ssim(self.gray_stream_image, gray_screenshot, full=False)
            return score > self.SSIM_DIFF_THRESHOLD
        except ValueError:
            return False



from typing import Any, Awaitable
from tapo import ApiClient, GenericDeviceHandler
from .AsyncDeviceState import AsyncDeviceState


class TapoState(AsyncDeviceState):
    """
    Class to represent the state of a generic Tapo device.
    """


    def __init__(self, ipv4: str, **kwargs) -> None:
        """
        Constructor.
        Initializes the Tapo device with its IP address,
        and creates a Tapo API client.

        Args:
            ipv4 (str): The Tapo device's IPv4 address.
            kwargs (dict): device-specific additional parameters,
                           including the TP-Link account credentials.
        """
        # Superclass constructor
        super().__init__(ipv4)

        # Tapo device connector
        self.client = ApiClient(
            kwargs.get("username", ""),
            kwargs.get("password", "")
        )
        self.device: GenericDeviceHandler = None
    

    async def _async_get_state(self) -> Awaitable[Any]:
        """
        Asynchronously get the status of the generic Tapo device.

        Returns:
            Awaitable[Any]: async function which returns the Tapo device's status.
        """
        if self.device is None:
            self.device = await self.client.generic_device(self.ipv4)

        info = await self.device.get_device_info()
        return info
    

    async def _async_close(self) -> None:
        """
        Asynchronously close the connection to the Tapo device.
        Simply sets the device and client to None.
        """
        self.device = None
        self.client = None



# Expose only concrete devices
from .devices import (
    init_device,
    close_device,

    # Plugs
    TpLinkPlug,
    TpLinkPlugTapo,
    TpLinkPlugSmartThings,
    TapoPlug,
    TapoPlugSmartThings,
    SmartThingsOutlet,
    TuyaPlug,
    
    # Lights
    TapoLight,
    TapoLightSmartThings,
    HueLight,
    HueLightEssentials,
    HueLightSmartThings,
    TuyaLight,

    # Cameras
    XiaomiCamera,
    TapoCamera,
    TapoCameraSmartThings,
    DLinkCamera

)



from ....DeviceState import ScreenshotState
from ....DeviceControl import TuyaControl, PlugControl


class TuyaPlug(ScreenshotState, TuyaControl, PlugControl):
    """
    Tuya generic plug.
    """

    ## Screen coordinates
    # Device controls screen coordinates
    device_x = 281
    device_y = 582
    # Toggle coordinates
    x = 540
    y = 1042



from .plugs import (
    TpLinkPlug,
    TpLinkPlugTapo,
    TpLinkPlugSmartThings,
    TapoPlug,
    TapoPlugSmartThings,
    SmartThingsOutlet,
    TuyaPlug
)

from .lights import (
    TapoLight,
    TapoLightSmartThings,
    HueLight,
    HueLightEssentials,
    HueLightSmartThings,
    TuyaLight
)

from .cameras import (
    XiaomiCamera,
    TapoCamera,
    TapoCameraSmartThings,
    DLinkCamera
)

from .device_lifecycle import init_device, close_device



from typing import Awaitable
from kasa import Device
from .AsyncDeviceState import AsyncDeviceState


class TpLinkPlugState(AsyncDeviceState):
    """
    Class to represent the state of a TP-Link plug.
    """

    def __init__(self, ipv4: str, **kwargs) -> None:
        """
        Constructor.

        Args:
            ipv4 (str): IPv4 address of the TP-Link plug.
            kwargs (dict): device-specific additional parameters.
        """
        # Superclass constructor
        super().__init__(ipv4)

        # TP-Link connector
        self.device_tplink: Device = None


    async def _async_get_state(self) -> Awaitable[bool]:
        """
        Asynchronously get the state of the TP-Link plug.

        Returns:
            bool: True if the plug is on, False otherwise.
        """
        if self.device_tplink is None:
            self.device_tplink = await Device.connect(host=self.ipv4)

        await self.device_tplink.update()
        return self.device_tplink.is_on
    

    async def _async_close(self) -> None:
        """
        Asynchronously close the connection to the TP-Link plug.
        """
        if self.device_tplink is not None:
            await self.device_tplink.disconnect()
            self.device_tplink = None



from typing import Any


class DeviceState:
    """
    Abstract class which represent a device state.
    """

    def __init__(self, ipv4: str, **kwargs) -> None:
        """
        Constructor.
        Initialize the device with its IP address.

        Args:
            ipv4 (str): The device's IPv4 address.
            kwargs (dict): device-specific additional parameters.
        """
        self.ipv4 = ipv4

        # Set the device-specific parameters
        for key, value in kwargs.items():
            setattr(self, key, value)


    def get_state(self) -> Any:
        """
        Get the state of the device.

        Returns:
            Any: current state of the device.
        """
        raise NotImplementedError


    def is_event_successful(self, previous_state: Any) -> bool:
        """
        Check if an event was successful,
        i.e. if the device is still reachable
        and its state has changed compared to the previous state.

        Args:
            previous_state (Any): The state of the device before the event.
        Returns:
            bool: True if the event was successful, False otherwise.
        """
        try:
            state = self.get_state()
        except Exception:
            return False
        else:
            return state != previous_state



from ....DeviceState import CameraScreenshotState
from ....DeviceControl import CameraControl


class DLinkCamera(CameraScreenshotState, CameraControl):
    """
    DLink camera (DCS-8000LH).
    """

    ### Class attributes
    # Android package
    android_package = "com.dlink.mydlinkunified"
    # SSIM threshold below which images are considered different
    SSIM_DIFF_THRESHOLD = 0.9
    ## Screen coordinates
    # Stream event
    x_start = 424.8
    y_start = 907.2
    x_stop = 57.6
    y_stop = 1008
    # Error message
    x_error = 360
    y_error = 806.4


    def start_stream(self) -> None:
        """
        Dismiss the potential error message,
        then start the camera's video stream.
        Overwrites the method from the parent class CameraControl.
        """
        phone = self.get_phone()
        phone.shell(f"input tap {self.x_error} {self.y_error}")
        phone.shell(f"input tap {self.x_start} {self.y_start}")



from ....DeviceState import HueState
from ....DeviceControl import SmartThingsLightControl


class HueLightSmartThings(HueState, SmartThingsLightControl):
    """
    Philips Hue light bulb,
    controlled by the SmartThings app.
    """

    ## Class variables
    # Screen coordinates to open the device controls
    device_x = 799
    device_y = 1139



from ....DeviceState import TpLinkPlugState
from ....DeviceControl import SmartThingsPlugControl


class TpLinkPlugSmartThings(TpLinkPlugState, SmartThingsPlugControl):
    """
    TP-Link HS110 plug,
    controlled through SmartThings.
    """

    ## Class attributes
    # Screen to open the device controls
    device_x = 281
    device_y = 776



from .DeviceState import DeviceState


class CameraState(DeviceState):
    """
    Abstract class for the state of a camera device.
    """


    def __init__(self, ipv4: str, **kwargs) -> None:
        """
        Constructor.
        Initialize the camera device with its IP address.

        Args:
            ipv4 (str): The device's IPv4 address.
            kwargs (dict): device-specific additional parameters.
        """
        self.ipv4 = ipv4
        self._was_last_stream_successful = False


    def is_event_successful(self, _) -> bool:
        """
        Check if the last stream event was successful.

        Returns:
            bool: True if the event was successful, False otherwise.
        """
        return self._was_last_stream_successful



from ....DeviceState import TpLinkPlugState
from ....DeviceControl import TapoPlugControl


class TpLinkPlugTapo(TpLinkPlugState, TapoPlugControl):
    """
    TP-Link HS110 plug,
    controlled through the Tapo app.
    """
    
    # Screen coordinates for device controls
    device_x = 799
    device_y = 654



from copy import deepcopy
from ppadb.device import Device as AdbDevice
from ppadb.client import Client as AdbClient
import cv2
import numpy as np


class DeviceControl:
    """
    Abstract class which provides methods to control a device.
    """

    ADB_SERVER_IP   = "127.0.0.1"
    ADB_SERVER_PORT = 5037
    adb_client = AdbClient(
        host=ADB_SERVER_IP,
        port=ADB_SERVER_PORT)


    def get_phone(self) -> AdbDevice:
        """
        Get the adb device object.

        Returns:
            ppadb.device.Device: The adb device object.
        Raises:
            IndexError: If no adb device is found.
        """
        phone = deepcopy(DeviceControl.adb_client.devices()[0])
        return phone


    def start_app(self) -> None:
        """
        Start the device's app on the phone.

        Raises:
            IndexError: If no adb device is found.
        """
        phone = self.get_phone()
        phone.shell(f"monkey -p {self.android_package} -c android.intent.category.LAUNCHER 1")
        del phone


    def close_app(self) -> None:
        """
        Close the device's app on the device.

        Raises:
            IndexError: If no adb device is found.
        """
        phone = self.get_phone()
        phone.shell(f"am force-stop {self.android_package}")
        del phone

    
    def screenshot(self) -> np.ndarray:
        """
        Take a screenshot of the device's screen.
        
        Returns:
            numpy.ndarray: The screenshot as a numpy array.
        Raises:
            IndexError: If no adb device is found.
        """
        phone = self.get_phone()
        screenshot_bytes = phone.screencap()
        screenshot_array = np.frombuffer(screenshot_bytes, dtype=np.uint8)
        image = cv2.imdecode(screenshot_array, cv2.IMREAD_UNCHANGED)
        del phone
        return image



from typing import Any, Awaitable
import threading
import asyncio
from .DeviceState import DeviceState


class AsyncDeviceState(DeviceState):
    """
    Asynchronous querying of a device's state.
    """


    def __init__(self, ipv4: str) -> None:
        """
        Constructor.
        Initialize the device with its IP address and start the asyncio event loop.

        Args:
            ipv4 (str): The device's IPv4 address.
        """
        # Device's IPv4 address
        self.ipv4 = ipv4

        # asyncio event loop
        self.loop = asyncio.new_event_loop()
        self.thread = threading.Thread(target=self._start_loop, daemon=True)
        self.thread.start()


    def _start_loop(self):
        """
        Start the asyncio event loop, and run it forever.
        """
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()


    def _async_get_state(self) -> Awaitable[Any]:
        """
        Asynchronously get the state of the device.
        This method is abstract and should be implemented in subclasses.

        Returns:
            Awaitable[Any]: async coroutine which returns the current state of the device.
        """
        raise NotImplementedError("This method should be implemented in subclasses.")


    def get_state(self) -> Any:
        """
        Get the device's state.

        Returns:
            Any: current state of the device.
        """
        return asyncio.run_coroutine_threadsafe(self._async_get_state(), self.loop).result()
    

    async def _async_close(self) -> None:
        """
        Asynchronously close the connection to the device.
        This method is abstract and should be implemented in subclasses.
        """
        raise NotImplementedError("This method should be implemented in subclasses.")


    def close(self) -> None:
        """
        Close the connection to the device.
        """
        asyncio.run_coroutine_threadsafe(self._async_close(), self.loop).result()
        self.loop.call_soon_threadsafe(self.loop.stop)
        self.thread.join()



from ....DeviceState import TpLinkPlugState
from ....DeviceControl import PlugControl


class TpLinkPlug(TpLinkPlugState, PlugControl):
    """
    TP-Link HS110 plug.
    """
    
    ## Class variables
    # Android package name
    android_package = "com.tplink.kasa_android"
    # Toggle coordinates
    x = 950
    y = 654



from .TpLinkPlug import TpLinkPlug
from .TpLinkPlugTapo import TpLinkPlugTapo
from .TpLinkPlugSmartThings import TpLinkPlugSmartThings
from .TapoPlug import TapoPlug
from .TapoPlugSmartThings import TapoPlugSmartThings
from .SmartThingsOutlet import SmartThingsOutlet
from .TuyaPlug import TuyaPlug



import numpy as np
import cv2
from skimage.metrics import structural_similarity as ssim
from .DeviceState import DeviceState


class ScreenshotState(DeviceState):
    """
    Device state using app screenshot.
    """

    # SSIM threshold below which images are considered different
    SSIM_DIFF_THRESHOLD = 0.95


    def get_state(self) -> np.ndarray:
        """
        Get the state of the device,
        by taking a screenshot of its app.

        Returns:
            numpy.ndarray: The screenshot as a numpy array.
        """
        return self.screenshot()
    

    def is_event_successful(self, previous_state: np.ndarray) -> bool:
        """
        Check if an event was successful,
        i.e. if the current state has changed compared to the given previous state.
        This method uses the SSIM metric to compare the screenshots.

        Args:
            previous_state (numpy.ndarray): app screenshot, as a numpy array, which represents the previous state.
        Returns:
            bool: True if the event was successful, False otherwise.
        """
        current_state = self.get_state()

        # Load images, and convert to grayscale
        gray_previous  = cv2.cvtColor(previous_state, cv2.COLOR_BGR2GRAY)
        gray_current   = cv2.cvtColor(current_state, cv2.COLOR_BGR2GRAY)

        # Compute SSIM
        score = ssim(gray_previous, gray_current, full=False)
        return score < self.SSIM_DIFF_THRESHOLD



import time
from ....DeviceState import CameraScreenshotState
from ....DeviceControl import CameraControl, SmartThingsControl


class TapoCameraSmartThings(CameraScreenshotState, CameraControl, SmartThingsControl):
    """
    Tapo camera (C200),
    controlled through the SmartThings app.
    """

    ### Class attributes
    # SSIM threshold below which images are considered different
    SSIM_DIFF_THRESHOLD = 0.975
    ## Screen coordinates
    # Device controls button
    device_x = 532.8
    device_y = 792
    # Expand stream
    x_expand = 525.6
    y_expand = 561.6
    # Toggle stream
    x_start = 720
    y_start = 360


    def start_app(self) -> None:
        """
        Start the SmartThings app on the phone,
        and open the camera stream.

        Raises:
            IndexError: If no adb device is found.
        """
        phone = self.get_phone()
        # Start SmartThings app
        phone.shell(f"monkey -p {SmartThingsControl.android_package} -c android.intent.category.LAUNCHER 1")
        time.sleep(10)
        # Open "Devices" tab
        phone.shell(f"input tap {SmartThingsControl.devices_x} {SmartThingsControl.devices_y}")
        time.sleep(5)
        # Open device controls
        phone.shell(f"input tap {self.device_x} {self.device_y}")
        time.sleep(15)
        # Expand stream to fullscreen
        phone.shell(f"input tap {self.x_expand} {self.y_expand}")
        phone.shell(f"input tap {self.x_expand} {self.y_expand}")
        time.sleep(3)
        # Pause stream
        phone.shell(f"input tap {self.x_start} {self.y_start}")
    

    def do_stop_stream(self) -> None:
        """
        Stop the camera's video stream.
        """
        phone = self.get_phone()
        phone.shell(f"input tap {self.x_start} {self.y_start}")
        phone.shell(f"input tap {self.x_start} {self.y_start}")



from enum import Enum
from .DeviceState import DeviceState


class LightState(DeviceState):
    """
    Abstract class for the state of a light device.
    """

    class StateKeys(Enum):
        """
        Enum for the keys in the dictionary representing the state of a light device.
        """
        IS_ON      = "is_on"
        BRIGHTNESS = "brightness"
        HUE        = "hue"
        SATURATION = "saturation"


    def is_off(self) -> bool:
        """
        Check if the light is on.
        
        Returns:
            bool: True if the light is on, False otherwise.
        """
        return not self.get_state()[self.StateKeys.IS_ON]



from ....DeviceState import TapoPlugState
from ....DeviceControl import SmartThingsPlugControl


class TapoPlugSmartThings(TapoPlugState, SmartThingsPlugControl):
    """
    Tapo P110 plug,
    controlled by the SmartThings app.
    """

    ## Class attributes
    # Screen coordinates to open the device controls
    device_x = 799
    device_y = 800



from .DeviceControl import DeviceControl
from .PlugControl import PlugControl
from .LightControl import LightControl
from .CameraControl import CameraControl
from .TapoControl import TapoControl
from .TapoPlugControl import TapoPlugControl
from .SmartThingsControl import SmartThingsControl
from .SmartThingsPlugControl import SmartThingsPlugControl
from .SmartThingsLightControl import SmartThingsLightControl
from .TuyaControl import TuyaControl



# Generic
from .DeviceState import DeviceState
from .AsyncDeviceState import AsyncDeviceState
from .ScreenshotState import ScreenshotState
from .LightState import LightState
from .CameraState import CameraState
from .CameraScreenshotState import CameraScreenshotState
from .TapoState import TapoState
from .SmartThingsState import SmartThingsState

# Plugs
from .TpLinkPlugState import TpLinkPlugState
from .TapoPlugState import TapoPlugState

# Lights
from .HueState import HueState
from .TapoLightState import TapoLightState



from .TapoLight import TapoLight
from .TapoLightSmartThings import TapoLightSmartThings
from .HueLight import HueLight
from .HueLightEssentials import HueLightEssentials
from .HueLightSmartThings import HueLightSmartThings
from .TuyaLight import TuyaLight



from .SmartThingsControl import SmartThingsControl
from .LightControl import LightControl


class SmartThingsLightControl(SmartThingsControl, LightControl):
    """
    Abstract class to control a generic light bulb through SmartThings.
    """

    ## Screen event coordinates
    # Toggle
    x = 961
    y = 436
    # Brightness and color gauges
    x_left_gauge  = 100
    x_right_gauge = 1000
    # Brightness
    y_brightness = 751
    # Color
    y_color = 1479



from ....DeviceState import TapoLightState
from ....DeviceControl import SmartThingsLightControl


class TapoLightSmartThings(TapoLightState, SmartThingsLightControl):
    """
    Tapo light bulb,
    controlled by the SmartThings app.
    """

    ## Class variables
    # Screen coordinates to open the device controls
    device_x = 532.8
    device_y = 1051.2



from ....DeviceState import TapoPlugState
from ....DeviceControl import TapoPlugControl


class TapoPlug(TapoPlugState, TapoPlugControl):
    """
    Tapo P110 plug.
    """
    
    # Screen coordinates for device controls
    device_x = 281
    device_y = 654



import time
from .DeviceControl import DeviceControl


class SmartThingsControl(DeviceControl):
    """
    Abstract class to control a generic device through SmartThings.
    """

    ## Class variables
    # Android package name
    android_package = "com.samsung.android.oneconnect"
    # Devices tab button screen coordinates
    devices_x = 335
    devices_y = 2230


    def start_app(self) -> None:
        """
        Start the SmartThings app on the phone,
        and open the device controls.

        Raises:
            IndexError: If no adb device is found.
        """
        phone = self.get_phone()
        # Start SmartThings app
        phone.shell(f"monkey -p {SmartThingsControl.android_package} -c android.intent.category.LAUNCHER 1")
        time.sleep(10)
        # Open "Devices" tab
        phone.shell(f"input tap {SmartThingsControl.devices_x} {SmartThingsControl.devices_y}")
        time.sleep(5)
        # Open device controls
        phone.shell(f"input tap {self.device_x} {self.device_y}")



from typing import Awaitable
from pysmartthings import Capability, Attribute
from ....DeviceState import SmartThingsState
from ....DeviceControl import SmartThingsPlugControl


class SmartThingsOutlet(SmartThingsState, SmartThingsPlugControl):
    """
    SmartThings Outlet.
    """

    ## Class attributes
    # Screen to open the device controls from the "Devices" tab
    device_x = 281
    device_y = 1139
    # Toggle coordinates from the device controls screen
    x = 454
    y = 1018
    # States boolean values
    states = {
        "on":  True,
        "off": False
    }


    async def _async_get_state(self) -> Awaitable[bool]:
        """
        Asynchronously get the state of the plug.

        Returns:
            Awaitable[bool]: async function which returns True if the plug is on, False otherwise.
        """
        status = await super()._async_get_state()
        state = status["main"][Capability.SWITCH][Attribute.SWITCH].value
        return SmartThingsOutlet.states[state]



import time
from ....DeviceState import HueState
from ....DeviceControl import LightControl


class HueLight(HueState, LightControl):
    """
    Philips Hue light bulb.
    """

    ### Class variables
    # Android package
    android_package = "com.philips.lighting.hue2"
    ## Screen coordinates
    # Zone button
    x_zone = 540
    y_zone = 679
    # Toggle button
    x = 162
    y = 1430


    def start_app(self) -> None:
        """
        Start the Philips Hue app on the phone,
        and open the light control screen.

        Raises:
            IndexError: If no adb device is found.
        """
        phone = self.get_phone()
        # Open app
        phone.shell(f"monkey -p {self.android_package} -c android.intent.category.LAUNCHER 1")
        time.sleep(10)
        # Open zone controls
        phone.shell(f"input tap {self.x_zone} {self.y_zone}")


    def set_brightness(self) -> None:
        raise NotImplementedError("Brightness cannot be set randomly for Philips Hue lights controlled through the Hue app.")
    

    def set_color(self) -> None:
        raise NotImplementedError("Color cannot be set randomly for Philips Hue lights controlled through the Hue app.")





from typing import List, Any, Awaitable
import asyncio
import aiohttp
from pysmartthings import SmartThings, Device
from .AsyncDeviceState import AsyncDeviceState


class SmartThingsState(AsyncDeviceState):
    """
    Abstract class which represents a generic SmartThings device's state.
    """


    @staticmethod
    async def _get_all_devices(token: str) -> Awaitable[List[Device]]:
        """
        Asynchronously get all the devices connected to SmartThings.

        Args:
            token (str): SmartThings API token.
        Returns:
            Awaitable[List]: async function which returns the list of all devices.
        """
        async with aiohttp.ClientSession() as session:
            api = SmartThings(session, token)
            devices = await api.get_devices()
            return devices
        
    
    @classmethod
    def get_all_devices(cls) -> List[Device]:
        """
        Get all the devices connected to SmartThings.

        Returns:
            List: list of all devices
        """
        return asyncio.run(cls._get_all_devices())
    

    @classmethod
    def print_all_devices(cls) -> None:
        """
        Print the label, name and id of all devices connected to SmartThings.
        """
        for device in cls.get_all_devices():
            print(device.label, device.name, device.device_id)

    
    def __init__(self, ipv4: str, **kwargs) -> None:
        """
        Constructor.

        Args:
            ipv4 (str): IPv4 address of the SmartThings device.
            kwargs (dict): device-specific additional parameters,
                           including SmartThings API token.
        """
        # Superclass constructor
        super().__init__(ipv4)

        # SmartThings instance
        self.id = kwargs.get("id", "")
        self.token = kwargs.get("token", "")
        self.http_session: aiohttp.ClientSession = None
        self.smartthings: SmartThings = None

    
    async def _async_get_state(self) -> Awaitable[Any]:
        """
        Asynchronously get the SmartThings device status.

        Returns:
            Awaitable[Any]: async function which returns the device status.
        """
        if self.http_session is None and self.smartthings is None:
            self.http_session = aiohttp.ClientSession()
            self.smartthings = SmartThings(session=self.http_session)
            self.smartthings.authenticate(self.token)
        
        status = await self.smartthings.get_device_status(self.id)
        return status
    

    async def _async_close(self) -> None:
        """
        Asynchronously close the SmartThings session.
        """
        if self.http_session is not None:
            await self.http_session.close()
            self.http_session = None
            self.smartthings  = None



from typing import Awaitable
from enum import Enum
import contextlib
import asyncio
import threading
from aiohue import HueBridgeV2, create_app_key
from .LightState import LightState


class HueState(LightState):
    """
    Class to represent the state of a Philips Hue light bulb.
    """


    class StateKeys(Enum):
        """
        Enum for the keys in the dictionary representing the state of the Philips Hue device.
        """
        IS_ON      = "is_on"
        BRIGHTNESS = "brightness"
        COLOR_X    = "color_x"
        COLOR_Y    = "color_y"

    
    @staticmethod
    async def _get_app_key(bridge_ip: str) -> Awaitable[str]:
        """
        Asynchronously get the app key for the Philips Hue bridge.

        Args:
            bridge_ip (str): IP address of the Philips Hue bridge.
        Returns:
            Awaitable[str]: async function which returns the app key.
        """
        input("Press the link button on the bridge and press enter to continue...")
        return await create_app_key(bridge_ip, "authentication_example")
    

    @classmethod
    def get_app_key(cls, bridge_ip: str) -> str:
        """
        Get the app key for the Philips Hue bridge.

        Args:
            bridge_ip (str): IP address of the Philips Hue bridge.
        Returns:
            str: app key.
        """
        with contextlib.suppress(KeyboardInterrupt):
            asyncio.run(cls._get_app_key(bridge_ip))

    
    def __init__(self, ipv4: str, **kwargs) -> None:
        """
        Constructor.

        Args:
            ipv4 (str): IPv4 address of the Philips Hue bridge.
            kwargs (dict): device-specific additional parameters,
                           including Philips Hue app key.
        """
        # Device's IPv4 address
        self.ipv4 = ipv4

        # App key for the Philips Hue bridge
        self.appkey = kwargs.get("appkey", "")

        # Hue bridge instance
        self.hue_bridge: HueBridgeV2 = None

        # asyncio event loop
        self.loop = asyncio.new_event_loop()
        self.thread = threading.Thread(target=self._start_loop, daemon=True)
        self.thread.start()


    def _start_loop(self):
        """
        Start the asyncio event loop, and run it forever.
        """
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()

    
    async def _async_get_state(self) -> Awaitable[dict]:
        """
        Get the state of the Philips Hue light asynchronously.

        Returns:
            Awaitable[dict]: async function which returns the status of the Philips Hue light.
        """
        if self.hue_bridge is None:
            self.hue_bridge = HueBridgeV2(self.ipv4, self.appkey)
            await self.hue_bridge.initialize()

        light = next(l for l in self.hue_bridge.lights.items)
        return {
            self.StateKeys.IS_ON: light.is_on,
            self.StateKeys.BRIGHTNESS: light.brightness,
            self.StateKeys.COLOR_X: light.color.xy.x,
            self.StateKeys.COLOR_Y: light.color.xy.y
        }
        
    
    def get_state(self) -> dict:
        """
        Get the state of the Philips Hue light.

        Returns:
            dict: The status of the Philips Hue light.
        """
        return asyncio.run_coroutine_threadsafe(self._async_get_state(), self.loop).result()
    

    async def _async_close(self) -> None:
        """
        Close the Philips Hue bridge connection asynchronously.
        """
        if self.hue_bridge is not None:
            await self.hue_bridge.close()
            self.hue_bridge = None
    

    def close(self) -> None:
        """
        Stop the Philips Hue light.
        """
        asyncio.run_coroutine_threadsafe(self._async_close(), self.loop).result()
        self.loop.call_soon_threadsafe(self.loop.stop)
        self.thread.join()



from typing import Awaitable
from .TapoState import TapoState
from .LightState import LightState


class TapoLightState(TapoState, LightState):
    """
    Class to represent the state of a Tapo light bulb.
    """


    async def _async_get_state(self) -> Awaitable[dict]:
        """
        Get the state of the light asynchronously.

        Returns:
            Awaitable[dict]: async function which returns the state of the light.
        """
        dev = await self.client.l530(self.ipv4)
        info = await dev.get_device_info()
        return {
            self.StateKeys.IS_ON:      info.device_on,
            self.StateKeys.BRIGHTNESS: info.brightness,
            self.StateKeys.HUE:        info.hue,
            self.StateKeys.SATURATION: info.saturation
        }



import time
import math
import random
from ....DeviceState import HueState
from ....DeviceControl import LightControl


class HueLightEssentials(HueState, LightControl):
    """
    Philips Hue light bulb,
    controlled by the Hue Essentials app.
    """

    ### Class variables
    # Android package
    android_package = "com.superthomaslab.hueessentials"
    ## Screen coordinates
    # Zone button
    x_zone = 540
    y_zone = 582
    # Lamp button
    x_lamp = 896
    y_lamp = 2036
    # Toggle button
    x = 950
    y = 679
    # Brightness
    x_left_gauge = 100
    x_right_gauge = 1000
    y_brightness = 388
    # Color
    x_color_center = 540
    y_color_center = 1382
    y_color_top    = 900


    def start_app(self, open_lamp_control: bool = False) -> None:
        """
        Start the Hue Essentials app on the phone,
        and open the light control screen.

        Args:
            open_lamp_control (bool): Whether to additionally open the lamp control screen after having opened the app.
                                      Optional, default is False.
        Raises:
            IndexError: If no adb device is found.
        """
        phone = self.get_phone()
        # Open app
        phone.shell(f"monkey -p {self.android_package} -c android.intent.category.LAUNCHER 1")
        time.sleep(10)
        # Open zone controls
        phone.shell(f"input tap {self.x_zone} {self.y_zone}")
        # If needed, open lamp controls
        if open_lamp_control:
            time.sleep(3)
            phone.shell(f"input tap {self.x_lamp} {self.y_lamp}")
    

    def do_set_color(self) -> None:
        """
        Randomly set the color of the Philips Hue light
        through the Hue Essentials app.
        """
        # Compute color wheel's radius
        radius = self.y_color_center - self.y_color_top
        # Generate random position on the color wheel
        distance  = random.randint(0, radius)
        angle_deg = random.randint(0, 360)
        angle_rad = angle_deg * math.pi / 180
        x = self.x_color_center + distance * math.cos(angle_rad)
        y = self.y_color_center + distance * math.sin(angle_rad)
        # Set the color
        self.get_phone().shell(f"input tap {x} {y}")



from ....DeviceState import TapoLightState
from ....DeviceControl import TapoControl, LightControl


class TapoLight(TapoLightState, TapoControl, LightControl):
    """
    Tapo L530E light.
    """

    ## Class variables
    # Device controls coordinates
    device_x = 187.2
    device_y = 705.6
    # Toggle coordinates
    x = 612
    y = 619.2
    # Brightness and color gauges
    x_left_gauge = 103
    x_right_gauge = 640
    # Brightness
    y_brightness = 1108.8
    # Color
    x_color_tab = 230.4
    y_color_tab = 1238.4
    y_color = 1310.4


    def do_set_color(self) -> None:
        """
        Ensure the "Color" tab is selected,
        then randomly set the color of the light. 
        """
        self.get_phone().shell(f"input tap {self.x_color_tab} {self.y_color_tab}")
        self._set_light_attr(self.y_color)



import time
from .DeviceControl import DeviceControl


class CameraControl(DeviceControl):
    """
    Class to control a camera device.
    """

    # Stream event duration
    stream_duration = 7


    def stream(self) -> None:
        """
        Perform the `stream` event.
        """
        self.start_stream()
        time.sleep(self.stream_duration)
        self.stop_stream()


    def start_stream(self) -> None:
        """
        Start the camera's video stream.
        """
        self.get_phone().shell(f"input tap {self.x_start} {self.y_start}")

    
    def do_stop_stream(self) -> None:
        """
        Stop the camera's video stream.
        """
        self.get_phone().shell(f"input tap {self.x_stop} {self.y_stop}")

    
    def stop_stream(self) -> None:
        """
        Check if the stream was successful,
        then stop the stream.
        """
        # Check if stream was successful
        self._was_last_stream_successful = self.get_state()
        
        if self._was_last_stream_successful:
            self.do_stop_stream()
